# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
'''Test arguments.'''

import sys
from inspect import getmembers, isfunction, signature

from docstring_parser import parse

from argufy import Argument

module = sys.modules[__name__]


def argument_type_hints_simple(check: int) -> None:
    '''Example demonstrating minimal CLI.'''
    pass


def test_argument_type_hints_simple() -> None:
    '''Test simple argument.'''
    name, fn = [
        x
        for x in getmembers(module, isfunction)
        if x[0] == 'argument_type_hints_simple'
    ][0]
    sig = signature(fn)
    docstring = parse(fn.__doc__)
    arguments = []
    for arg in sig.parameters:
        document = next(
            (d for d in docstring.params if d.arg_name == arg), None
        )
        arguments.append(
            Argument(parameters=sig.parameters[arg], docstring=document)
        )
    # print('Arguments: ', arguments[0].__dict__)
    assert not hasattr(arguments[0], 'default')
    # assert arguments[0].metavar == 'INT'


def argument_type_hints_bool(
    check_false: bool = False, check_true: bool = True
) -> None:
    '''Example bool.'''
    pass


def test_argument_type_hints_bool() -> None:
    '''Test simple boolean.'''
    name, fn = [
        x
        for x in getmembers(module, isfunction)
        if x[0] == 'argument_type_hints_bool'
    ][0]
    sig = signature(fn)
    docstring = parse(fn.__doc__)
    arguments = []
    for arg in sig.parameters:
        document = next(
            (d for d in docstring.params if d.arg_name == arg), None
        )
        arguments.append(
            Argument(parameters=sig.parameters[arg], docstring=document)
        )
    # print(arguments[0].__dict__)
    assert arguments[0].default is False
    assert arguments[1].default is True


def argument_type_hints_all(
    string_check: str = 'A',
    bool_check: bool = False,
    integer_check: int = 1,
    float_check: float = 1.5,
    list_check: list = ['A'],
    set_check: set = {'a'},
    tuple_check: tuple = ('A',),
    # file_check: open = 'test.toml',
) -> None:
    '''Example full.'''
    pass


def test_argument_type_hints_all() -> None:
    '''Test full type set.'''
    name, fn = [
        x
        for x in getmembers(module, isfunction)
        if x[0] == 'argument_type_hints_all'
    ][0]
    sig = signature(fn)
    document = parse(fn.__doc__)
    arguments = []
    for arg in sig.parameters:
        docstring = next(
            (d for d in document.params if d.arg_name == arg), None
        )
        arguments.append(
            Argument(parameters=sig.parameters[arg], docstring=docstring)
        )
    # print(arguments[0].__dict__)
    assert arguments[0].type == str
    assert arguments[0].default == 'A'

    # print(arguments[1].__dict__)
    # NOTE: Argparse does not store bool type
    assert not hasattr(arguments[1], 'type')
    assert arguments[1].default is False

    # print(arguments[2].__dict__)
    assert arguments[2].type == int
    assert arguments[2].default == 1

    # print(arguments[3].__dict__)
    assert arguments[3].type == float
    assert arguments[3].default == 1.5

    # print(arguments[4].__dict__)
    # assert arguments[4].type == list
    assert arguments[4].default == ['A']

    # print(arguments[5].__dict__)
    # assert arguments[5].type == set
    assert arguments[5].default == {'a'}

    # print(arguments[6].__dict__)
    # assert arguments[6].type == tuple
    assert arguments[6].default == ('A',)



"""Simple argparse."""

# from pprint import pprint
import argparse
from argparse import Action, HelpFormatter
from typing import Iterable, Optional

import colorama
from colorama import Fore, Style

colorama.init()


class ArgufyHelpFormatter(HelpFormatter):
    """Provide formatting for Argufy."""

    # argparse.HelpFormatter(prog, max_help_position=80, width=130)

    def add_usage(
        self,
        usage: Optional[str],
        actions: Iterable[Action],
        groups: Iterable[argparse._ArgumentGroup],
        prefix: Optional[str] = 'usage: ',
    ) -> None:
        """Format usage message."""
        if prefix is not None:
            prefix = self.font(prefix)
        super(ArgufyHelpFormatter, self).add_usage(
            usage, actions, groups, prefix
        )

    @staticmethod
    def font(text: str, width: str = 'BRIGHT') -> str:
        """Set the string thickness."""
        return getattr(Style, width) + text + Style.RESET_ALL

    @staticmethod
    def shade(text: str, color: str = 'CYAN') -> str:
        """Set the string color."""
        return getattr(Fore, color.upper()) + text + Style.RESET_ALL

    # def _format_action_invocation(self, action: Action) -> str:
    #     """Format arguments summary."""
    #     # TODO: find alternative that does not modify action
    #     if isinstance(action, argparse._SubParsersAction):
    #         if action.choices is not None:
    #             for choice in list(action.choices):
    #                 parser = action.choices.pop(choice)
    #                 choice = self.shade(choice)
    #                 action.choices[choice] = parser
    #     return super(
    #         ArgufyHelpFormatter, self
    #     )._format_action_invocation(action)

    def _expand_help(self, action: Action) -> str:
        """Format help message."""
        if action.help:
            return self.shade(
                super(ArgufyHelpFormatter, self)
                ._expand_help(action)
                .rstrip('.')
                .lower(),
                'YELLOW',
            )
        else:
            return ''

    def _format_action(self, action: Action) -> str:
        """Format arguments."""
        if isinstance(action, argparse._SubParsersAction._ChoicesPseudoAction):
            subcommand = self.shade(
                self.font(self._format_action_invocation(action))
            )
            help_text = self._expand_help(action)
            # TODO: calculate correct spacing
            return f"    {subcommand.ljust(37)}{help_text}\n"
        else:
            # action.option_strings = [
            #     self.font(self.shade(option))
            #     for option in action.option_strings
            # ]
            return super(ArgufyHelpFormatter, self)._format_action(action)



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: Apache 2.0, see LICENSE for more details.
'''Provide documentation tasks.'''

from invoke import Context, task


@task
def lint(ctx):  # type: (Context) -> None
    '''Check code for documentation errors.'''
    ctx.run('pydocstyle')


@task
def coverage(ctx):  # type: (Context) -> None
    '''Ensure all code is documented.'''
    ctx.run('docstr-coverage **/*.py')


@task(pre=[lint], post=[coverage])
def test(ctx):  # type: (Context) -> None
    '''Test documentation build.'''
    ctx.run('mkdocs build')


@task
def build(ctx):  # type: (Context) -> None
    '''Build documentation site.'''
    ctx.run('mkdocs build')


@task
def publish(ctx):  # type: (Context) -> None
    '''Publish project documentation.'''
    ctx.run('mkdocs gh-deploy')



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: Apache 2.0, see LICENSE for more details.
'''Initialize project management tasks.'''

from invoke import Collection

from . import build, doc, qa

ns = Collection().from_module(build)
ns.add_collection(Collection.from_module(doc))
ns.add_collection(Collection.from_module(qa))



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
# type: ignore
'''Test parser.

Attributes
----------
check_attribute: bool
    Check document attributes

'''

from ast import literal_eval
import sys

import pytest

from argufy import __version__, Parser

sys.path.append('.')
import subcommands_parser  # noqa: E402


def test_help():
    '''Do help function for CLI.'''
    # NOTE: getting varying results
    # 0 displays help
    # 2 dumps error (invalid choice)

    parser = Parser(__version__, command_type='subcommand')
    parser.add_commands(subcommands_parser, exclude_prefixes=['test_'])

    with pytest.raises(SystemExit) as blank_err:
        parser.dispatch([])
    assert blank_err.type == SystemExit

    with pytest.raises(SystemExit) as help_err:
        parser.dispatch(['--help'])
    assert help_err.type == SystemExit
    assert help_err.type == blank_err.type

    with pytest.raises(SystemExit) as err:
        parser.dispatch(['subcommands-parser', 'example-bool', '--help'])
    assert err.type == SystemExit
    assert err.value.code == 0


def test_bool(capsys):
    '''Do main function for CLI.'''
    parser = Parser(command_type='subcommand')
    parser.add_commands(subcommands_parser, exclude_prefixes=['test_'])
    parser.dispatch([
        'subcommands-parser',
        'example-bool',
        '--bool-check',
    ])
    capture = capsys.readouterr()
    assert literal_eval(capture.out) is True


def test_choice(capsys):
    '''Do main function for CLI.'''
    parser = Parser(command_type='subcommand')
    parser.add_commands(subcommands_parser, exclude_prefixes=['test_'])
    parser.dispatch([
        'subcommands-parser',
        'example-choice',
        '--choice-check',
        'B',
    ])
    capture = capsys.readouterr()
    assert literal_eval(capture.out) is True



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
"""Test parser one.

Attributes
----------
example_variable1: str
    first example variable for testing
example_variable2: str
    second example variable for testing

"""
example_variable1 = 'test1'


def _example_hidden(check: str = 'hidden-result') -> None:
    """Demonstrate example hidden command.

    Parameters
    ----------
    check: str
        example variable for hidden command

    """
    print(check)


def example_bool(bool_check: bool = False) -> None:
    """Demonstrate example bool.

    Parameters
    ----------
    bool_check: bool, optional
        example boolean

    """
    print(bool_check)


def example_choice(choice_check: str = 'A') -> None:
    """Demonstrate example choice.

    Parameters
    ----------
    choice_check: str, {'A', 'B', 'C'}
        example choice

    """
    print(choice_check)



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: Apache 2.0, see LICENSE for more details.
"""Provide CLI for example."""

import sys
from argufy import Parser

from . import cmd1, cmd2


def main() -> None:
    """Demonstrate main with CLI."""
    parser = Parser(
        # command_type='subcommand',
        version='0.1.0',
        use_module_args=True,
        main_args_builder={
            'module': 'examples.multiparse.cmd2',
            'function': 'builder',
            'instance': '_settings',
            'variables': {
                'var1': 'yey'
            }
        },
        log_level='debug',
        log_handler=sys.stderr,
    )
    parser.add_commands(cmd1)
    parser.add_commands(cmd2)
    parser.dispatch()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
"""Test parser two.

Attributes
----------
example_variable: str
    Example variable for testing

"""

from typing import List, Tuple

example_variable2 = 'test2'


def example_list(list_check: List[str] = []) -> None:
    """Demonstrate example bool.

    Parameters
    ----------
    list_check: list, optional
        example boolean

    """
    print(list_check)


def example_tuple(tuple_check: Tuple[str, ...] = ('A', 'B', 'C')) -> None:
    """Demonstrate example choice.

    Parameters
    ----------
    tuple_check: tuple, ('A', 'B', 'C')
        example choice

    """
    print(tuple_check)



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
# type: ignore

'''Test parser.

Attributes
----------
check_attribute: bool
    Check document attributes

'''

from ast import literal_eval

import pytest
import sys

from argufy import __version__, Parser

sys.path.append('.')
import command_parser  # noqa: E402
# module = sys.modules[__name__]


def test_help():
    '''Do help function for CLI.'''
    parser = Parser(version=__version__)
    parser.add_commands(command_parser, exclude_prefixes=['test_'])
    with pytest.raises(SystemExit) as blank_err:
        parser.dispatch([])
    assert blank_err.type == SystemExit
    with pytest.raises(SystemExit) as help_err:
        parser.dispatch(['--help'])
    assert help_err.type == SystemExit
    assert blank_err.value.code == help_err.value.code


def test_bool(capsys):
    '''Do main function for CLI.'''
    parser = Parser()
    parser.add_commands(command_parser, exclude_prefixes=['test_'])
    parser.dispatch(['example-bool', '--bool-check'])
    capture = capsys.readouterr()
    assert literal_eval(capture.out) is True


def test_choice(capsys):
    '''Do main function for CLI.'''
    parser = Parser()
    parser.add_commands(command_parser, exclude_prefixes=['test_'])
    parser.dispatch(['example-choice', '--choice-check', 'B'])
    capture = capsys.readouterr()
    assert literal_eval(capture.out) is True



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
"""Test parser."""

from . import builder

_settings = builder()


def example_builder() -> None:
    """Demonstrate variable imported from config."""
    print('---', _settings, '---')


def example_args(*args_check: str) -> None:
    """Demonstrate example arguments.

    Parameters
    ----------
    args_check: str
        example arguments

    """
    print(args_check)


def example_kwargs(**kwargs: str) -> None:
    """Demonstrate example keyword arguments.

    Parameters
    ----------
    variable_one: str
        variable one

    variable_test: str
        variable two

    variable_three: str
        variable three

    """
    if kwargs.get('variable_one'):
        print(kwargs['variable_one'])

    if kwargs.get('variable_two'):
        print(kwargs['variable_two'])

    if kwargs.get('variable_three'):
        print(kwargs['variable_three'])



"""Provide example to update variable."""

from dataclasses import dataclass


@dataclass
class Settings:
    """Provide settings example using dataclasses."""

    var1: str
    var2: str = 'default'



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: Apache 2.0, see LICENSE for more details.
"""Provide CLI for example."""

# import sys

from argufy import Parser

from . import cmd1, cmd2


def main() -> None:
    """Demonstrate main with CLI."""
    parser = Parser(
        command_type='subcommand',
        # log_level='debug',
        # log_handler=sys.stderr,
    )
    parser.add_commands(cmd1)
    parser.add_commands(cmd2)
    parser.dispatch()


if __name__ == '__main__':
    main()



"""Simple argparse."""

import logging
import sys
from typing import Optional

from argufy import Parser

module = sys.modules[__name__]

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def optional_boolean(
    check: str,
    bool_check: Optional[bool] = None,
) -> None:
    """Run example bool."""
    print('check is:', check)

    if bool_check:
        print(f"tis {bool_check}")
    else:
        print('nothing to see here')


parser = Parser(
    prog='complete',
    version='1.2.3',
    log_level='DEBUG',
    log_handler=sys.stderr,
    # command_scheme='chain',
)
parser.add_commands(module)
parser.dispatch()



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
"""Test parser.

Attributes
----------
example_variable: str
    Example variable for testing

"""

import json  # noqa
from typing import Optional

example_variable = 'ex_var'
optional_variable: Optional[str] = None


def print_variable() -> None:
    """Print example variable."""
    print('example_varible', example_variable)
    print('optional_varible', optional_variable)


def example_bool(bool_check: bool = False) -> None:
    """Demonstrate example bool.

    Parameters
    ----------
    bool_check: bool, optional
        example boolean

    """
    print(bool_check)


def example_choice(choice_check: str = 'A') -> None:
    """Demonstrate example choice.

    Parameters
    ----------
    choice_check: str, {'A', 'B', 'C'}
        example choice

    """
    print(choice_check)



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: Apache 2.0, see LICENSE for more details.
"""Provide package."""

from argparse import ArgumentError, ArgumentTypeError


class ArgufyException(ArgumentError):
    """Provide exception for parser errors."""


class ArgufyTypeError(ArgumentTypeError):
    """Provide exception for argument type errors."""



# -*- coding: utf-8 -*-
# type: ignore
# copyright: (c) 2020 by Jesse Johnson.
# license: Apache 2.0, see LICENSE for more details.
'''Test Task-Runner.'''

from typing import Optional, TYPE_CHECKING

from invoke import call, task

from argufy import __version__

if TYPE_CHECKING:
    from invoke import Context

if (
    'dev' in __version__
    or 'a' in __version__
    or 'b' in __version__
    or 'rc' in __version__
):
    part = 'build'
else:
    part = 'patch'


@task
def build(ctx, format=None):  # type: (Context, Optional[bool]) -> None
    '''Build wheel package.'''
    if format:
        ctx.run(f"flit build --format={format}")
    else:
        ctx.run('flit build')


@task(pre=[call(build, format='wheel')])
def dev(ctx):  # type: (Context) -> None
    '''Perform development runtime environment setup.'''
    ctx.run('flit install --pth-file --python python3')


@task
def install(
    ctx,
    symlink=True,
    dev=False,
):  # type: (Context, bool, bool) -> None
    '''Install within environment.'''
    args = []
    if symlink:
        args.append('--symlink')
    if dev:
        args.append('--python=python3')
    ctx.run(f"flit install {' '.join(args)}")


@task
def publish(ctx):  # type: (Context) -> None
    '''Publish project distribution.'''
    ctx.run('flit publish')


@task
def clean(ctx):  # type: (Context) -> None
    '''Clean project dependencies and build.'''
    paths = ['dist', 'logs']
    paths.append('**/__pycache__')
    paths.append('**/*.pyc')
    paths.append('argufy.egg-info')
    for path in paths:
        ctx.run(f"rm -rf {path}")


@task
def version(
    ctx,
    part=part,
    tag=False,
    commit=False,
    message=None,
):  # type: (Context, str, bool, bool, Optional[str]) -> None
    '''Update project version and apply tags.'''
    args = [part]
    if commit:
        args.append('--commit')
    else:
        args.append('--dry-run')
        args.append('--allow-dirty')
        args.append('--verbose')
        print('Add "--commit" to actually bump the version.')
    if tag or message:
        args.append('--tag')
        if message:
            args.append(f"--tag-message '{message}'")
    ctx.run(f"bumpversion {' '.join(args)}")



"""Simple argparse."""

import inspect
import logging
import sys
from typing import Optional

from argufy import Parser

module = sys.modules[__name__]

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def switch():  # type: ignore
    """Run empty function to check switch."""
    print('test empty switch')


def positional(test: str) -> None:
    """Run example positional.

    Parameters
    ----------
    test: str
        example test variable

    """
    print(test)


def boolean(bool_check=False):  # type: ignore
    """Run example bool."""
    print(bool_check)


def choice(choice_check='A'):  # type: ignore
    """Run example choice.

    Parameters
    ----------
    choice_check: str, {'A', 'B', 'C'}
        example choice

    """
    print(choice_check)


def optional(variable: Optional[str]) -> None:
    """Run example optional."""
    if variable:
        print(variable)
    else:
        print('yey! optional is not set')


def arguments(test_arg: str = 'test', *args: str, **kwargs: str) -> None:
    """Run example key arguments.

    Parameters
    ----------
    test_arg: str, optional
        test argument with value
    test1: str, optional
        kwargs test one
    test2: str, optional
        kwargs test two

    """
    print('test_arg', test_arg)
    if args != []:
        print('args', args)

    if kwargs != {}:
        print('kwargs', kwargs)


# frame = inspect.currentframe()
# args, args_paramname, kwargs_paramname, values = (
#     inspect.getargvalues(frame)
# )
# print(args, args_paramname, kwargs_paramname, values)
# for k, v in dict(values).items():
#     print(k, v)
# print(inspect.getfullargspec(arguments))
# for x in inspect.getmembers(arguments):
#     print(x)

sig = inspect.signature(arguments)
# print(sig.parameters['args'])
# print(sig.parameters['kwargs'])

parser = Parser(
    prog='complete',
    version='1.2.3',
    log_level='DEBUG',
    log_handler=sys.stderr,
    # command_scheme='chain',
)
parser.add_commands(module)
parser.dispatch()



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: Apache 2.0, see LICENSE for more details.
"""Arguments for inspection based CLI parser."""

import inspect
# import logging
import re
import typing
from ast import literal_eval
from typing import Any, List, Optional, Union

from docstring_parser.common import DocstringParam


class Argument:
    """Represent argparse arguments."""

    __short_flags: List[str] = ['-h']

    def __init__(
        self,
        docstring: DocstringParam,
        parameters: Optional[inspect.Parameter] = None,
    ) -> None:
        """Initialize argparse argument."""
        # self.attributes: Dict[Any, Any] = {}

        # set parameter default
        if parameters:
            self.default = parameters.default
            self.name = parameters  # type: ignore
        else:
            self.default = None

        # set parameter type
        if (
            parameters
            and parameters.annotation != inspect._empty  # type: ignore
        ):
            self.__parse_parameters(parameters)
        elif docstring and docstring.type_name:
            self.__parse_docstring(docstring)
        elif self.default is not None:
            self.type = type(self.default)

        # if hasattr(self, 'type'):
        #     self.metavar = (self.type.__name__)

        # set parameter help message
        if docstring:
            self.help = docstring.description

    def __parse_parameters(self, parameters: inspect.Parameter) -> None:
        """Get parameter types from type inspection."""
        # if typing.get_origin(parameters.annotation) is Union:
        # XXX need to handle types from typing
        # print(
        #     '---', parameters.annotation,
        #     typing.get_origin(parameters.annotation),
        # )
        if hasattr(parameters.annotation, '__origin__'):
            annotation = typing.get_args(parameters.annotation)
            # print('annotation', annotation, parameters.annotation)

            # check if annotation is optional
            if type(None) in annotation:
                self.nargs = '?'
            else:
                # self.type = annotation
                self.type = typing.get_origin(parameters.annotation)
        else:
            self.type = parameters.annotation

    def __parse_docstring(self, docstring: DocstringParam) -> None:
        """Get parameter types from docstring."""
        # Parse docstring for parameter types and defaults
        if ',' in docstring.type_name:
            for arg in docstring.type_name.split(',', 1):
                if not hasattr(self, 'type'):
                    # NOTE: Limit input that eval will parse
                    if arg.__class__.__module__ == 'builtins':
                        self.type = literal_eval(arg) if arg != 'str' else str
                if arg.lower() == 'optional' and not hasattr(self, 'default'):
                    # XXX: should optional not exist instead of 'None'
                    self.default = None
                # TODO: tighten regex
                if re.search(r'^\s*\{.*\}\s*$', arg):
                    self.choices = literal_eval(arg.strip())
        if not hasattr(self, 'type'):
            # NOTE: Limit input that eval will parse
            if docstring.type_name in (
                ('float', 'int', 'str', 'list', 'dict', 'tuple')
            ):
                self.type = eval(docstring.type_name)  # nosec

    @property
    def name(self) -> List[str]:
        """Get argparse command/argument name."""
        return self.__name

    @name.setter
    def name(self, parameters: inspect.Parameter) -> None:
        """Set argparse command/argument name."""
        name = parameters.name.replace('_', '-')

        # parse positional argument
        if (
            not hasattr(self, 'default')
            and not str(parameters).startswith('**')
        ):
            self.__name = [name]
            if str(parameters).startswith('*'):
                self.nargs = '*'

        # parse optional argument
        else:
            if str(parameters).startswith('**'):
                self.nargs = '*'
            flags = [f"--{name}"]

            # NOTE: check for conflicting flags
            if '-' not in name:
                # TODO: check if common short flag (ex: version)
                n = name[:1]
                if n not in Argument.__short_flags:
                    Argument.__short_flags.append(n)
                    flags.insert(0, f"-{n}")
                elif n.upper() not in Argument.__short_flags:
                    Argument.__short_flags.append(n.upper())
                    flags.insert(0, f"-{n.upper()}")
            self.__name = flags

    @property
    def type(self) -> Any:
        """Get argparse argument type."""
        return self.__type  # type: ignore

    @type.setter
    def type(self, annotation: Any) -> None:
        """Set argparse argument type."""
        # log.debug('prematched annotation:', annotation)
        # print(annotation)
        if annotation == bool:
            # NOTE: these store bool type internally
            if self.default or not hasattr(self, 'default'):
                self.action = 'store_false'
            else:
                self.action = 'store_true'
        elif annotation == int:
            self.__type = annotation
            self.action = 'append'
        elif annotation == list:
            self.__type = lambda v: int(v) if v.isdigit() else v
            self.action = 'append'
        elif annotation == tuple:
            self.__type = annotation
            self.nargs = '+'
        elif annotation == set:
            self.__type = annotation
            self.nargs = '+'
        else:
            # log.debug('unmatched annotation:', annotation)
            self.__type = annotation
            # self.nargs = 1

    @property
    def metavar(self) -> str:
        """Get argparse argument metavar."""
        return self.__metavar

    @metavar.setter
    def metavar(self, metavar: str) -> None:
        """Set argparse argument metavar."""
        # NOTE: Only positional arguments use metavars
        if not hasattr(self, 'default'):
            self.__metavar = metavar

    # @property
    # def const(self) -> str:
    #     """Get argparse argument const."""
    #     return self.__const

    # @const.setter
    # def const(self, const: str) -> None:
    #     """Set argparse argument const."""
    #     self.__const = const

    # @property
    # def dest(self) -> str:
    #     """Get argparse command/argument dest."""
    #     return self.__dest

    # @dest.setter
    # def dest(self, dest: str) -> None:
    #     """Set argparse command/argument dest."""
    #     self.__dest = dest

    # @property
    # def required(self) -> bool:
    #     """Get argparse required argument."""
    #     return self.__required

    # @required.setter
    # def required(self, required: bool) -> None:
    #     """Set argparse required argument."""
    #     self.__required = required

    @property
    def action(self) -> str:
        """Get argparse argument action."""
        return self.__action

    @action.setter
    def action(self, action: str) -> None:
        """Set argparse argument action."""
        self.__action = action

    @property
    def choices(self) -> List[str]:
        """Get argparse argument choices."""
        return self.__choices

    @choices.setter
    def choices(self, choices: set) -> None:
        """Set argparse argument choices."""
        self.__choices = list(choices)

    @property
    def nargs(self) -> Union[int, str]:
        """Get argparse argument nargs."""
        return self.__nargs

    @nargs.setter
    def nargs(self, nargs: Union[int, str]) -> None:
        """Set argparse argument nargs."""
        # TODO: map nargs to argparse with typing
        # 1: set number of values
        # ?: a single optional value
        # *: a flexible list of values
        # +: like * requiring at least one value
        # REMAINDER: unused args
        self.__nargs = nargs

    @property
    def default(self) -> Any:
        """Get argparse argument default."""
        return self.__default

    @default.setter
    def default(self, default: Any) -> None:
        """Set argparse argument default."""
        if default != inspect._empty:  # type: ignore
            self.__default = default
        # XXX: this keeps conflicting with positional arguments
        # else:
        #     self.__default = None

    @property
    def help(self) -> str:
        """Get argparse command/argument help message."""
        return self.__help

    @help.setter
    def help(self, description: str) -> None:
        """Set argparse command/argument help message."""
        self.__help = description



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: Apache 2.0, see LICENSE for more details.
"""Argufy is an inspection based CLI parser."""

import inspect
import logging
import sys
import typing
from argparse import ArgumentParser, Namespace, _SubParsersAction
# from dataclasses import is_dataclass
from inspect import Signature, _ParameterKind
from types import ModuleType
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

from docstring_parser import DocstringParam, parse as docstring_parse

from argufy.argument import Argument
from argufy.formatter import ArgufyHelpFormatter

log = logging.getLogger(__name__)

# Define function as parameters for MyPy
F = TypeVar('F', bound=Callable[..., Any])


class Parser(ArgumentParser):
    """Provide CLI parser for function."""

    exclude_prefixes = ('@', '_')

    def __init__(self, *args: str, **kwargs: Any) -> None:
        """Initialize parser.

        Parameters
        ----------
        prog: str
            The name of the program
        usage: str
            The string describing the program usage
        description: str
            Text to display before the argument help
        epilog: str
            Text to display after the argument help
        parents: list
            A list of ArgumentParser objects whose arguments should also
            be included
        formatter_class: Object
            A class for customizing the help output
        prefix_chars: char
            The set of characters that prefix optional arguments
        fromfile_prefix_chars: None
            The set of characters that prefix files from which additional
            arguments should be read
        argument_default: None
            The global default value for arguments
        conflict_handler: Object
            The strategy for resolving conflicting optionals
        add_help: str
            Add a -h/--help option to the parser
        allow_abbrev: bool
            Allows long options to be abbreviated if the abbreviation is
            unambiguous

        """
        # TODO: handle environment variables

        module = self.__get_parent_module()
        if module:
            docstring = docstring_parse(module.__doc__)
            if not kwargs.get('description'):
                kwargs['description'] = docstring.short_description

            if 'prog' not in kwargs:
                kwargs['prog'] = module.__name__.split('.')[0]

        if 'version' in kwargs:
            self.prog_version = kwargs.pop('version')

        # if 'prefix' in kwargs:
        #     self.prefix = kwargs.pop('prefix')
        # else:
        #     self.prefix = kwargs['prog'].upper()
        # log.debug(self.prefix)

        if 'log_level' in kwargs:
            log.setLevel(getattr(logging, kwargs.pop('log_level').upper()))
        if 'log_handler' in kwargs:
            log_handler = kwargs.pop('log_handler')
            log.addHandler(logging.StreamHandler(log_handler))

        self.use_module_args = kwargs.pop('use_module_args', False)
        self.main_args_builder = kwargs.pop('main_args_builder', None)
        self.command_type = kwargs.pop('command_type', None)
        self.command_scheme = kwargs.pop('command_scheme', None)

        if 'formatter_class' not in kwargs:
            self.formatter_class = ArgufyHelpFormatter

        super().__init__(**kwargs)

        # NOTE: cannot move to formatter
        self._positionals.title = ArgufyHelpFormatter.font(
            self._positionals.title or 'arguments'
        )
        self._optionals.title = ArgufyHelpFormatter.font(
            self._optionals.title or 'flags'
        )

        # XXX version lookup infinite loop when absent
        if hasattr(self, 'prog_version'):
            self.add_argument(
                '--version',
                action='version',
                version=f"%(prog)s {self.prog_version}",
                help='display application version',
            )

    @staticmethod
    def __get_parent_module() -> Optional[ModuleType]:
        """Get name of module importing this module."""
        stack = inspect.stack()
        # TODO: need way to better identify parent module
        stack_frame = stack[2]
        result = inspect.getmodule(stack_frame[0]) or None
        return result

    @staticmethod
    def __clean_args(argument: Argument) -> Dict[Any, Any]:
        """Retrieve cleaned parameters from an Argument."""
        return {
            k[len('_Argument__'):]: v
            for k, v in vars(argument).items()
            if k.startswith('_Argument__')
        }

    @staticmethod
    def _get_excludes(exclude_prefixes: tuple = tuple()) -> tuple:
        """Combine class excludes with instance."""
        if exclude_prefixes != []:
            return tuple(exclude_prefixes) + Parser.exclude_prefixes
        else:
            return Parser.exclude_prefixes

    @staticmethod
    def __get_description(
        name: str, docstring: DocstringParam,
    ) -> Optional[str]:
        """Get argument description from docstring."""
        return next((d for d in docstring.params if d.arg_name == name), None)

    @staticmethod
    def __get_keyword_args(
        signature: Signature, docstring: DocstringParam,
    ) -> List[str]:
        """Get keyward arguments from docstring."""
        parameters = [x for x in signature.parameters]
        return [
            x.arg_name
            for x in docstring.params
            if x.arg_name not in parameters
        ]

    @staticmethod
    def __generate_parameter(
        name: str, module: ModuleType,
    ) -> inspect.Parameter:
        """Generate inpect parameter."""
        parameter = inspect.Parameter(
            name,
            _ParameterKind.POSITIONAL_OR_KEYWORD,
            default=getattr(module, name),
            annotation=inspect._empty,  # type: ignore
        )
        return parameter

    def add_commands(
        self,
        module: ModuleType,
        parser: Optional[ArgumentParser] = None,
        exclude_prefixes: tuple = tuple(),
        command_type: Optional[str] = None,
    ) -> 'Parser':
        """Add commands.

        Parameters
        ----------
        module: ModuleType,
            Module used to import functions for CLI commands.
        parser: ArgumentParser, optional
            Parser used to append subparsers to create subcommands.
        exclude_prefixes: tuple,
            Methods from a module that should be excluded.
        command_type: str, optional
            Choose format type of commands to be created.

        Returns
        -------
        self:
            Return object itself to allow chaining functions.

        """
        # use self or an existing parser
        if not parser:
            parser = self
        parser.formatter_class = ArgufyHelpFormatter

        module_name = module.__name__.split('.')[-1]
        docstring = docstring_parse(module.__doc__)
        excludes = Parser._get_excludes(exclude_prefixes)

        # use exsiting subparser or create a new one
        if not any(isinstance(x, _SubParsersAction) for x in parser._actions):
            # TODO: use metavar for hidden commands
            parser.add_subparsers(dest=module_name, parser_class=Parser)

        # check if command exists
        command = next(
            (x for x in parser._actions if isinstance(x, _SubParsersAction)),
            None,
        )

        # set command name scheme
        if command_type is None:
            command_type = self.command_type

        # create subcommand for command
        if command_type == 'subcommand':
            if command:
                msg = docstring.short_description
                subcommand = command.add_parser(
                    module_name.replace('_', '-'),
                    description=msg,
                    formatter_class=self.formatter_class,
                    help=msg,
                )
                subcommand.set_defaults(mod=module)

            # append subcommand to exsiting command or create a new one
            return self.add_commands(
                module=module,
                parser=subcommand,
                exclude_prefixes=Parser._get_excludes(exclude_prefixes),
                command_type='command',
            )

        # TODO: separate into method
        for name, value in inspect.getmembers(module):
            # TODO: Possible singledispatch candidate
            if not name.startswith(excludes):
                # skip classes for now
                if inspect.isclass(value):
                    # TODO: check if dataclass instance
                    # TODO: check if class instance
                    continue  # pragma: no cover

                # create commands from functions
                elif inspect.isfunction(value):
                    # TODO: Turn parameter-less function into switch

                    # merge builder function maing_args into parser
                    if (
                        self.main_args_builder
                        and name == self.main_args_builder['function']
                    ):
                        self.add_arguments(value, parser)

                    # create commands from functions
                    elif (
                        module.__name__ == value.__module__
                        and not name.startswith(', '.join(excludes))
                        or (
                            self.main_args_builder
                            and name == self.main_args_builder['function']
                        )
                    ):
                        # create command from function
                        if command:
                            # control command name format
                            if self.command_scheme == 'chain':
                                cmd_name = f"{module_name}.{name}"
                            else:
                                cmd_name = name

                            msg = docstring_parse(value.__doc__)\
                                .short_description
                            cmd = command.add_parser(
                                cmd_name.replace('_', '-'),
                                description=msg,
                                formatter_class=self.formatter_class,
                                help=msg,
                            )
                            cmd.set_defaults(mod=module, fn=value)
                        # add arguments from function
                        # log.debug(f"command {name} {value} {cmd}")
                        self.add_arguments(value, cmd)

                # create arguments from module varibles
                elif (
                    self.use_module_args
                    and not isinstance(value, ModuleType)
                    and not hasattr(typing, name)
                    and (
                        self.main_args_builder
                        and name != self.main_args_builder['instance']
                    )
                ):
                    # TODO: Reconcile inspect parameters with dict
                    # TODO: use argparse.SUPPRESS for hidden arguments
                    arguments = self.__clean_args(
                        Argument(
                            self.__get_description(name, docstring),
                            self.__generate_parameter(name, module),
                        )
                    )
                    name = arguments.pop('name')
                    parser.add_argument(*name, **arguments)
        return self

    def add_arguments(
        self, obj: Any, parser: Optional[ArgumentParser] = None
    ) -> 'Parser':
        """Add arguments to parser/subparser.

        Parameters
        ----------
        obj: Any
            Verious module, function, or arguments that can be inspected.
        parser: ArgumentParser, optional
            Parser/Subparser that arguments will be added.

        Returns
        -------
        self:
            Return object itself to allow chaining functions.

        """
        if not parser:
            parser = self

        # prep object for inspection
        docstring = docstring_parse(obj.__doc__)
        signature = inspect.signature(obj)

        # populate subcommand with keyword arguments
        for arg in signature.parameters:
            param = signature.parameters[arg]
            description = self.__get_description(arg, docstring)
            # log.debug(f"param: {param}, {param.kind}")

            if not param.kind == inspect.Parameter.VAR_KEYWORD:
                log.debug(f"param annotation: {param.annotation}")
                argument = self.__clean_args(Argument(description, param))
                name = argument.pop('name')
                # print(name, argument)
                parser.add_argument(*name, **argument)

        # populate options
        # log.debug(f"params {params}")
        for arg in self.__get_keyword_args(signature, docstring):
            description = self.__get_description(arg, docstring)
            arguments = self.__clean_args(Argument(description))
            parser.add_argument(f"--{arg.replace('_', '-')}", **arguments)

        # log.debug(f"arguments {arguments}")
        # TODO for any docstring not collected parse here (args, kwargs)
        # log.debug('docstring params', docstring.params)
        return self

    def __set_main_arguments(self, ns: Namespace) -> Namespace:
        """Separate and set main arguments from builder function.

        Paramters
        ---------
        ns: Namespace
            Argparse namespace object for a command.

        Returns
        -------
        Namespace:
            Argparse namespace object with command arguments.

        """
        # pass main arguments to builder function
        if self.main_args_builder:
            builder_mod = sys.modules[self.main_args_builder['module']]
            builder = getattr(builder_mod, self.main_args_builder['function'])
            builder_signature = inspect.signature(builder)
            builder_args = {}
            for param in builder_signature.parameters:
                if param in vars(ns):
                    builder_args[param] = vars(ns).pop(param)
            builder_mod.__dict__[
                self.main_args_builder['instance']
            ] = builder(**builder_args)
        return ns

    def __set_module_arguments(
        self, fn: Callable[[F], F], ns: Namespace
    ) -> Namespace:
        """Separate and set module arguments from functions.

        Paramters
        ---------
        fn: Callable
            Function used to seperate module arguments from function.
        ns: Namespace
            Argparse namespace object for a command.

        Returns
        -------
        Namespace:
            Argparse namespace object with command arguments.

        """
        # XXX: only works on subcommands that use 'mod'
        if 'mod' in ns:
            mod = vars(ns).pop('mod')
        else:
            mod = None

        # separate namespace from other variables
        signature = inspect.signature(fn)
        docstring = docstring_parse(fn.__doc__)

        # inspect non-signature keyword args
        keywords = self.__get_keyword_args(signature, docstring)
        args = [
            {k: vars(ns).pop(k)}
            for k in list(vars(ns).keys()).copy()
            if not signature.parameters.get(k) and k not in keywords
        ]
        log.debug(f"arguments {args}, {keywords}")

        # set module variables
        if mod and self.use_module_args:
            for arg in args:
                for k, v in arg.items():
                    mod.__dict__[k] = v
        return ns

    def retrieve(
        self,
        args: Sequence[str] = sys.argv[1:],
        ns: Optional[Namespace] = None,
    ) -> Tuple[List[str], Namespace]:
        """Retrieve parsed values from CLI input.

        Paramters
        ---------
        args: Sequence[str]
            Command line arguments passed to the parser.
        ns: Optional[Namespace]
            Argparse namespace object for a command.

        Returns
        -------
        List[str]:
            Argparse remaining unparse arguments.
        Namespace:
            Argparse namespace object with command arguments.

        """
        # TODO: handle invalid argument

        # show help when no arguments provided
        if args == []:
            args = ['--help']  # pragma: no cover
        main_ns, main_args = self.parse_known_args(args, ns)
        if main_args == [] and 'fn' in vars(main_ns):
            return main_args, main_ns
        else:
            # default to help message for subcommand
            if 'mod' in vars(main_ns):
                a = []
                a.append(vars(main_ns)['mod'].__name__.split('.')[-1])
                a.append('--help')
                self.parse_args(a)
            return main_args, main_ns

    def dispatch(
        self,
        args: Sequence[str] = sys.argv[1:],
        ns: Optional[Namespace] = None,
    ) -> Optional[Callable[[F], F]]:
        """Call command with arguments.

        Paramters
        ---------
        args: Sequence[str]
            Command line arguments passed to the parser.
        ns: Optional[Namespace]
            Argparse namespace object for a command.

        Returns
        -------
        Optional[Callable[[F], F]]:
            Call function with arguments.

        """
        # parse variables
        arguments, namespace = self.retrieve(args, ns)
        log.debug(f"dispatch: {arguments}, {namespace}")

        main_ns_result = self.__set_main_arguments(namespace)

        # call function with variables
        if 'fn' in namespace:
            ns_vars = vars(namespace)
            fn = ns_vars.pop('fn')

            self.__set_module_arguments(fn, main_ns_result)

            # XXX: only takes standard types
            # attempt to plug parameters using inspect
            splat = None
            signature = inspect.signature(fn)
            for arg in signature.parameters:
                param = signature.parameters[arg]
                if (
                    str(param).startswith('*')
                    and not str(param).startswith('**')
                ):
                    splat = ns_vars.pop(arg)

            # XXX: only works with splat and kwargs
            if splat:
                fn(*splat, **ns_vars)
            else:
                fn(**ns_vars)
        return self.dispatch(arguments) if arguments != [] else None



"""Provide prototype for group design."""

import argparse

# from argufy import Parser

# parent_parser = Parser()
parent_parser = argparse.ArgumentParser(add_help=False)
parent_parser.add_argument('--parent', type=int)

# foo_parser = Parser(parents=[parent_parser])
foo_parser = argparse.ArgumentParser(parents=[parent_parser])
foo_parser.add_argument('foo')
print(foo_parser.parse_args(['--parent', '2', 'XXX']))

# bar_parser = Parser(parents=[parent_parser])
bar_parser = argparse.ArgumentParser(parents=[parent_parser])
bar_parser.add_argument('--bar')
print(bar_parser.parse_args(['--bar', 'YYY']))

# print(parent_parser.parse_args())



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
"""Test parser.

Attributes
----------
variable: str
    Example variable for testing

"""

from .config import Settings

variable = 'ex_var'


# def print_variable() -> None:
#     """Print example variable."""
#     print('example_varible', variable)


def example_settings(settings: Settings, clear: bool = False) -> None:
    """Demonstrate example settings."""
    # Parameters
    # ----------
    # bool_check: settings
    #     example boolean

    # """
    print(settings, clear)



#!/usr/bin/python3
"""Simple argparse."""

import sys

from argufy import Parser

module = sys.modules[__name__]


def switch():  # type: ignore
    """Empty function to check switch."""
    print('test empty switch')


def positional(test: str):  # type: ignore
    """Run example positional."""
    print(test)


def example_bool(bool_check=False):  # type: ignore
    """Run example bool."""
    print(bool_check)


def example_choice(choice_check='A'):  # type: ignore
    """Run example choice."""
    print(choice_check)


parser = Parser(version='1.2.3')
parser.add_commands(module)
parser.dispatch()



"""Provide initialization for multiparse."""

from .config import Settings


def builder(var1: str = 'test1', var2: str = 'test2') -> Settings:
    """Demonstrate variable imported from config.

    Parameters
    ----------
    var1: str
        Variable one.
    var2: str
        Variable two.

    """
    return Settings(var1=var1, var2=var2)



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
# type: ignore
'''Test restructured arguments.'''

import sys
from inspect import getmembers, isfunction, signature

from docstring_parser import parse

from argufy import Argument

module = sys.modules[__name__]


def argument_restructured_bool(check_false=False, check_true=True):
    '''Mock restructured example bool.

    :param bool_check: list packages and v1ersion
    :type bool_check: bool

    '''
    pass


def test_argument_restructured_bool():
    '''Test restructured example boolean.'''
    name, fn = [
        x
        for x in getmembers(module, isfunction)
        if x[0] == 'argument_restructured_bool'
    ][0]
    sig = signature(fn)
    docstring = parse(fn.__doc__)
    arguments = []
    for arg in sig.parameters:
        document = next(
            (d for d in docstring.params if d.arg_name == arg), None
        )
        arguments.append(
            Argument(parameters=sig.parameters[arg], docstring=document)
        )
    # print(arguments[0].__dict__)
    assert arguments[0].default is False
    assert arguments[1].default is True


def argument_restructured_choice(choice='A'):
    '''Mock restructured example choice.

    {'A','B','C'}

    :param choice: argument choice
    :type str: choice

    '''
    pass


def test_argument_restructured_choice():
    '''Test restructured choices.'''
    name, fn = [
        x
        for x in getmembers(module, isfunction)
        if x[0] == 'argument_restructured_choice'
    ][0]
    sig = signature(fn)
    document = parse(fn.__doc__)
    arguments = []
    for arg in sig.parameters:
        docstring = next(
            (d for d in document.params if d.arg_name == arg), None
        )
        arguments.append(
            Argument(parameters=sig.parameters[arg], docstring=docstring)
        )
    # print(arguments[0].__dict__)
    assert arguments[0].help == 'argument choice'
    assert arguments[0].default == 'A'


def argument_restructured_all(
    string_check='A',
    bool_check=False,
    integer_check=1,
    float_check=1.5,
    list_check=['A'],
    set_check={'a'},
    tuple_check=('A',),
    # file_check='test.toml',
):
    '''Mock restructured full type set.

    :param string_check: argument string
    :type string_check: str
    :param bool_check: argument bool
    :type bool_check: bool
    :param integer_check: argument int
    :type integer_check: int
    :param float_check: argument float
    :type float_check: float
    :param list_check: argument list
    :type list_check: list
    :param set_check: argument set
    :type set_check: set
    :param tuple_check: argument tuple
    :type tuple_check: tuple

    '''
    pass


def test_argument_restructured_all():
    '''Test restructured full type set.'''
    name, fn = [
        x
        for x in getmembers(module, isfunction)
        if x[0] == 'argument_restructured_all'
    ][0]
    sig = signature(fn)
    document = parse(fn.__doc__)
    arguments = []
    for arg in sig.parameters:
        docstring = next(
            (d for d in document.params if d.arg_name == arg), None
        )
        arguments.append(
            Argument(parameters=sig.parameters[arg], docstring=docstring)
        )
    # print(arguments[0].__dict__)
    assert arguments[0].help == 'argument string'
    # assert arguments[0].type == str
    assert arguments[0].default == 'A'

    # print(arguments[1].__dict__)
    assert arguments[1].help == 'argument bool'
    # NOTE: Argparse does not store bool type
    # assert not hasattr(arguments[1], 'type')
    assert arguments[1].default is False

    # print(arguments[2].__dict__)
    assert arguments[2].help == 'argument int'
    # assert arguments[2].type == int
    assert arguments[2].default == 1

    # print(arguments[3].__dict__)
    assert arguments[3].help == 'argument float'
    # assert arguments[3].type == float
    assert arguments[3].default == 1.5

    # print(arguments[4].__dict__)
    assert arguments[4].help == 'argument list'
    # assert arguments[4].type == list
    assert arguments[4].default == ['A']

    # print(arguments[5].__dict__)
    assert arguments[5].help == 'argument set'
    # assert arguments[5].type == set
    assert arguments[5].default == {'a'}

    # print(arguments[6].__dict__)
    assert arguments[6].help == 'argument tuple'
    # assert arguments[6].type == tuple
    assert arguments[6].default == ('A',)



"""Simple argparse."""

import logging
import sys

from argufy import Parser

module = sys.modules[__name__]

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def find(*args: str, **kwargs: str) -> None:
    """Run example key arguments.

    Parameters
    ----------
    test_one: str, optional
        kwargs test one
    test_two: str, optional
        kwargs test two

    """
    if args != []:
        print('args', args)

    if kwargs != {}:
        print('kwargs', kwargs)


def ignore(test: str) -> None:
    """Ignore this."""
    print('this should be ignored:', test)


parser = Parser(
    prog='complete',
    version='1.2.3',
    # log_level='debug',
    # log_handler=sys.stderr,
)
parser.add_commands(module)
parser.dispatch()



"""Provide example to update variable."""

from dataclasses import dataclass


@dataclass
class Settings:
    """Provide example configuration settings.

    Parameters
    ----------
    var1: str
        Variable one.
    var2: str
        variable two.

    """

    var1: str
    var2: str



# -*- coding: utf-8 -*-
# type: ignore
# copyright: (c) 2020 by Jesse Johnson.
# license: Apache 2.0, see LICENSE for more details.
'''Test Task-Runner.'''

from typing import Optional, TYPE_CHECKING

from invoke import task

if TYPE_CHECKING:
    from invoke import Context


@task
def style(ctx, check=True):  # type: (Context, bool) -> None
    '''Format project source code to PEP-8 standard.'''
    args = ['--skip-string-normalization']
    if check:
        args.append('--check')
    ctx.run('isort --atomic **/*.py')
    ctx.run(f"black **/*.py {' '.join(args)}")


@task
def lint(ctx):  # type: (Context) -> None
    '''Check project source code for linting errors.'''
    ctx.run('flake8')


@task
def type_check(ctx, path='.'):  # type: (Context, str) -> None
    '''Check project source types.'''
    ctx.run(f"mypy {path}")


@task
def unit_test(ctx, capture=None):  # type: (Context, Optional[str]) -> None
    '''Perform unit tests.'''
    args = []
    if capture:
        args.append(f"--capture={capture}")
    ctx.run(f"pytest {' '.join(args)}")


@task
def static_analysis(ctx):  # type: (Context) -> None
    '''Perform static code analysis on imports.'''
    ctx.run('safety check')
    ctx.run('bandit -r argufy')


@task
def coverage(ctx, report=None):  # type: (Context, Optional[str]) -> None
    '''Perform coverage checks for tests.'''
    args = ['--cov=argufy']
    if report:
        args.append(f"--cov-report={report}")
    ctx.run(f"pytest {' '.join(args)} ./tests/")


@task(pre=[style, lint, unit_test, static_analysis, coverage])
def test(ctx):
    '''Run all tests.'''
    pass



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
'''Test version.'''

from argufy import __version__


def test_version() -> None:
    '''Test project version is managed.'''
    assert __version__ == "0.1.2b1"



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
# type: ignore
'''Test parser.

Attributes
----------
check_attribute: bool
    Check document attributes

'''

# check_attribute = 'test-attr'


def example_bool(bool_check: bool = False):
    '''Mock example bool.

    Parameters
    ----------
    bool_check: bool, optional
        list packages and version

    '''
    print(bool_check is True)


def example_choice(choice_check: str = 'A'):
    '''Mock example choice.

    Parameters
    ----------
    choice_check: str, {'A', 'B', 'C'}
        example choice

    '''
    print(choice_check == 'B')



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
"""Inspection based parser based on argparse."""

import logging
from typing import List

from argufy.argument import Argument  # noqa
from argufy.formatter import ArgufyHelpFormatter  # noqa
from argufy.parser import Parser  # noqa

__author__ = 'Jesse P. Johnson'
__author_email__ = 'jpj6652@gmail.com'
__title__ = 'argufy'
__description__ = 'Inspection based parser built on argparse.'
__version__ = '0.1.2b1'
__license__ = 'Apache-2.0'
__copyright__ = 'Copyright 2020 Jesse Johnson.'
__all__: List[str] = ['Argument', 'ArgufyHelpFormatter', 'Parser']

logging.getLogger(__name__).addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-
# :copyright: (c) 2020 by Jesse Johnson.
# :license: Apache 2.0, see LICENSE for more details.
"""Test parser.

Attributes
----------
example_variable: str
    Example variable for testing

"""


def example_args(*args_check: str) -> None:
    """Demonstrate example arguments.

    Parameters
    ----------
    args_check: str
        example arguments

    """
    print(args_check)


def example_kwargs(**kwargs_check: str) -> None:
    """Demonstrate example keyword arguments.

    Parameters
    ----------
    kwargs_check: str, {'A': 'a', 'B': 'b', 'C': 'c'}
        example keyword arguments

    """
    print(kwargs_check)



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: Apache 2.0, see LICENSE for more details.
"""Provide CLI for example."""

import sys
from argufy import Parser

from . import cmd


def main() -> None:
    """Demonstrate main with CLI."""
    parser = Parser(
        # command_type='subcommand',
        use_module_args=True,
        version='0.1.0',
        log_level='debug',
        log_handler=sys.stderr,
    )
    parser.add_commands(cmd)
    parser.dispatch()


if __name__ == '__main__':
    main()



"""Example argument group."""

# import argparse
# import inspect
# import traceback

from argufy import Parser

parser = Parser()
# parser = argparse.ArgumentParser(description='Foo', add_help=False)

group1 = parser.add_argument_group('group1')
# group1.add_argument('test-g1')
group1.add_argument('-1', '--option-one', help='get option one')

group2 = parser.add_argument_group('group2')
# group2.add_argument('test-g2')
group2.add_argument('-2', '--option-two', help='get option two')

print(parser.parse_args())
# parser.dispatch()
